import re as _regex
from collections.abc import Iterable as _Iterable


def snake_to_camel(s: str, always_uppercase: _Iterable[str] = ('id',), unless_str_in_always_uppercase: bool = False) -> str:
  if unless_str_in_always_uppercase and s.lower() in always_uppercase:
    return s

  words = s.split('_')
  converted = words[0] + ''.join(word.title() for word in words[1:])

  for word in always_uppercase:
    converted = converted.replace(word.title(), word.upper())

  return converted



class _Getch:
  """Get a single character from standard input.

  Warning: May behave differently on Unix and Windows
  """

  def __init__(self):
    try:
      self.impl = _GetchWindows()
    except ImportError:
      self.impl = _GetchUnix()

  def __call__(self):
    """Get a single character from standard input.

    Warning: May behave differently on Unix and Windows
    """
    return self.impl()


class _GetchUnix:
  def __init__(self):
    import sys
    import termios
    import tty

    self.sys = sys
    self.termios = termios
    self.tty = tty

  def __call__(self):
    fd = self.sys.stdin.fileno()
    old_settings = self.termios.tcgetattr(fd)
    try:
      self.tty.setraw(self.sys.stdin.fileno())
      ch = self.sys.stdin.read(1)
    finally:
      self.termios.tcsetattr(fd, self.termios.TCSADRAIN, old_settings)
    return ch.encode(encoding='ascii')


class _GetchWindows:
  def __init__(self):
    import msvcrt

    self.msvcrt = msvcrt

  def __call__(self):
    return self.msvcrt.getch()


getch = _Getch()



import pathlib as p

from .discord.tcrd_string import get_token

# Get the test.py file contents if any and execute them.
TESTFILE = get_token(filename='test.py', depth=0, dont_strip=True)
exec(TESTFILE)



import os
from collections.abc import Callable
from functools import wraps


def _send(s: str) -> None:
  """Internal printhook."""
  print(s, end='')


class Cursor:
  def hide(self, *, return_code_only: bool = False) -> None | str:
    a = '\033[?25l'
    if return_code_only:
      return a
    else:
      _send(a)

  def unhide(self, *, return_code_only: bool = False) -> None | str:
    a = '\033[?25h'
    if return_code_only:
      return a
    else:
      _send(a)


class _TerminalSizeTuple(tuple):
  def __new__(cls, width: int, height: int):
    return super().__new__(cls, (width, height))

  def __init__(self, width: int, height: int) -> None:
    self.width = width
    self.height = height

  def __complex__(self):
    return complex(self.width, self.height)

  def __str__(self):
    return f'{self.width}x{self.height}'


class Terminal:
  @property
  def width(self):
    return os.get_terminal_size().columns

  @property
  def height(self):
    return os.get_terminal_size().lines

  @property
  def size(self):
    return _TerminalSizeTuple(*os.get_terminal_size())

  def hide_cursor_during_function(self, func: Callable) -> Callable:
    def wrapper(*args, **kwargs):
      self.cursor.hide()
      r = func(*args, **kwargs)
      self.cursor.unhide()
      return r

    return wrapper

  cursor = Cursor()


terminal = Terminal()

__all__ = ['terminal']



from collections.abc import Callable
from functools import wraps
from typing import Any, NoReturn

from .tcr_void import void


class MayNotReturn:
  """Denote a return value for a function that may or may not return."""


if True:  # \/ # @trei

  def trei(
    exception: BaseException = Exception,
    excepth: Callable = void,
    els: Callable = void,
    finaly: Callable = void,
  ) -> Callable:
    if not issubclass(exception, BaseException):
      msg = f'exception must be an instance of BaseException (got {type(exception)}, {exception})'
      raise TypeError(msg)

    def wrap(func: Callable, exception=exception, excepth=excepth, els=els, finaly=finaly):
      @wraps(func)
      def wrapper(*args, **kwargs):
        try:
          res = func(*args, **kwargs)
        except exception as e:
          excepth(e, *args, **kwargs)
        else:
          els(res, *args, **kwargs)
        finally:
          finaly(*args, **kwargs)

      return wrapper

    return wrap


if True:  # \/ # asert

  def asert(condition_func: Callable, errmsg: str | None = None, *args: Any, **kwargs: Any) -> Any | MayNotReturn:
    if not condition_func(*args, **kwargs):
      if errmsg:
        raise AssertionError(errmsg)
      raise AssertionError



import os
import pathlib as p
import re as regex
import string
from typing import Any, Literal, NoReturn

import arc

from ..src.tcr_compare import able
from ..src.tcr_constants import BACKTICKS
from ..src.tcr_null import UniqueDefault as RaiseError
from .tcrd_snowflake import is_snowflake


def get_token(filename: str = 'TOKEN.txt', depth=2, *, dont_strip=False, default=RaiseError) -> None | str:
  """Get the nearest file with name=filename (default 'TOKEN.txt') and return its stripped contents (unless specified not to strip with `dont_strip=True`).

  This algoritm searches for files named TOKEN.txt (or custom name) in the current directory, then the parent directory, then the parent of parent and so on.
  The search continues up to depth `depth` (depth=0: current directory only, depth=1, current directory and its parent, etc.). If multiple files are found, return the closest one to the current directory.
  If no file is found, an error will be raised unless `default` is provided.
  """

  def rexit(x):
    os.chdir(origin_path)
    return x

  origin_path = p.Path.cwd().absolute()

  os.chdir(origin_path)

  f = origin_path / filename

  for i in range(depth + 1):
    f = (origin_path / '/'.join(['..'] * i)) / filename
    if f.is_file():
      t = f.read_text()
      if not dont_strip:
        t = t.strip()
      return rexit(t)

  if default is not RaiseError:
    return default

  raise FileNotFoundError(f'Unable to locate token file: {filename}')


def _ify_snowflake_errorer(id: Any) -> None:
  if not is_snowflake(id, allow_string=True):
    if isinstance(id, int | str):
      err = ValueError('id is not a valid snowflake, use tcr.discord.is_snowflake() to validate if needed.')
    else:
      err = TypeError(f'Expected str or int, got {type(id)} instead.')

    raise err


class IFYs:
  """Features related to turning IDs into user/channel/command/etc. mentions, emojis, and more if i can think of any.

  https://discord.com/developers/docs/reference#message-formatting

  Example:
  ```py
  >>> IFYs.userify(1234)
  '<@1234>'
  ```

  Includes argument validation:
  - Raises TypeError if invalid type was supplied (not str or int)
  - Raises ValueError if the type is valid, but snowflake in it is not (for example if negative int was passed in)
  """

  @staticmethod
  def userify(user_id: int | str) -> str:
    """### User mentions.

    ```py
    >>> IFYs.userify(1234)
    '<@1234>'
    ```
    """

    _ify_snowflake_errorer(user_id)

    return f'<@{user_id}>'

  @staticmethod
  def userbangify(user_id: int | str) -> str:
    """### User mentions (Old discord system).

    ```py
    >>> IFYs.userbangify(1234)
    '<@!1234>'
    ```
    """

    _ify_snowflake_errorer(user_id)

    return f'<@!{user_id}>'

  @staticmethod
  def channelify(channel_id: int | str) -> str:
    """### Channel mentions.

    ```py
    >>> IFYs.channelify(1234)
    '<#1234>'
    ```
    """

    _ify_snowflake_errorer(channel_id)

    return f'<#{channel_id}>'

  @staticmethod
  def roleify(role_id: int | str) -> str:
    """### Role mentions.

    ```py
    >>> IFYs.roleify(1234)
    '<@&1234>'
    ```
    """

    _ify_snowflake_errorer(role_id)

    return f'<@&{role_id}>'

  @staticmethod
  def commandify(command_name: str, command_id: int | str) -> str:
    """### Command mentions.

    I'm not sure how to validate `command_name` other than the type (python regex apparently doesn't support whatever discord wrote [here](https://discord.com/developers/docs/interactions/application-commands#application-command-object-application-command-naming))

    ```py
    >>> IFYs.commandify('settings', 1234)
    '</settings:1234>'
    ```
    """

    _ify_snowflake_errorer(command_id)

    if not isinstance(command_name, str):
      raise TypeError(f'Expected str, got {type(command_name)} instead.')

    return f'</{command_name}:{command_id}>'

  @staticmethod
  def emojify(emoji_name: str, emoji_id: int | str, *, animated: bool = False) -> str:
    """### Emojis (animated and non-animated).

    I'm not sure how to validate `emoji_name` other than the type (i'm not aware of discord's emoji naming scheme unlime command_name where they at least published some regex? maybe they did somewhere but if they did as i said im not aware of it).

    ```py
    >>> IFYs.emojify('mmLol', 1234)
    '<:mmLol:1234>'
    >>> IFYs.emojify('b1nzy', 1234, animated=True)
    '<a:b1nzy:1234>'
    ```
    """

    _ify_snowflake_errorer(emoji_id)

    if not isinstance(emoji_name, str):
      raise TypeError(f'Expected str, got {type(emoji_name)} instead.')

    return f'<{"a" if animated else ""}:{emoji_name}:{emoji_id}>'

  @staticmethod
  def timeify(timestamp: int | str, style: None | Literal['t', 'T', 'd', 'D', 'f', 'F', 'R'] = None) -> str:
    if not able(int, timestamp):
      raise TypeError(f'Expected intable object, got {f"{type(timestamp)} (but a str that does not contain an int)" if isinstance(timestamp, str) else type(timestamp)} instead.')

    if style not in (None, 't', 'T', 'd', 'D', 'f', 'F', 'R'):
      raise ValueError(f'Expected style=None, or a valid format specifier (https://discord.com/developers/docs/reference#message-formatting-timestamp-styles)')

    return f'<t:{timestamp}{":" if style is not None else ""}{style if style is not None else ""}>'

  @staticmethod
  def specialify(option: Literal['customize', 'browse', 'guide'] = 'customize') -> str:
    if option not in ('customize', 'browse', 'guide'):
      raise ValueError(f"Expected option in ('customize', 'browse', 'guide'), got {option} instead.")

    return f'<id:{option}>'

  @staticmethod
  def ctx_commandify(ctx: arc.Context) -> str:
    _ify_snowflake_errorer(ctx.interaction.command_id)

    return f'<{ctx.command.display_name}:{ctx.interaction.command_id}>'



import asyncio as aio
from collections.abc import Callable
from collections.abc import Coroutine as Korwine
from typing import Any


class RunSACAble(Callable, Korwine): ...


async def run_sac(func_or_coro: RunSACAble, *args, **kwargs) -> Any:
  """Run any function, async function or coroutine and return the result."""
  if aio.iscoroutine(func_or_coro):
    if args or kwargs:
      raise ValueError('Invalid arguments when using with coro')
    return await func_or_coro

  if aio.iscoroutinefunction(func_or_coro):
    return await func_or_coro(*args, **kwargs)

  if callable(func_or_coro):
    return func_or_coro(*args, **kwargs)

  msg = 'func_or_coro must be a callable or coroutine.'
  raise TypeError(msg)



class DiscordLimits:
  class Message:
    LENGTH = 4000
    LENGTH_SAFE = 1984  # Slightly less than 2000 for safety & to fit in a reference hehe
    LENGTH_SAFEST = 1800
    FILE_SIZE_MB = 25

  class Embed:
    TITLE = 256
    DESCRIPTION = 4096
    DESCRIPTION_SAFE = 4000
    AUTHOR_NAME = 256

    class Fields:
      AMOUNT = 25
      TITLE = 256
      DESCRIPTION = 1024

    FOOTER = 2048
    TOTAL_CHARACTERS = 6000



import inspect
import sys
from typing import TypeVar

T = TypeVar('T')


def get_lineno(*, default: T | None = None, backtrack_frames: int = 1) -> int | T:
  """Return the python file's line number at which this function was called as an int.

  Args:
    default: Any = None, will be returned if the function's calling place could not be determiend
    backtrack_frames: int = 1, how many frames to go back in the stack, if set to 1, will return the calling frame's line number, if set to 0 will return this module's internal line number (so dont) but may be useful when building a function that itself makes use of get_lineno() in which case you'd need to make it 2 so your function doesn't just return internal line number
  """
  frame = inspect.currentframe()
  for _ in range(backtrack_frames):
    if frame is None:
      return default
    frame = frame.f_back
  return frame.f_lineno if frame else default


def get_file_colon_lineno(default: T | None = None, backtrack_frames: int = 1, additional_offset: int = 0) -> str:
  """Return the python file's line number at which this function was called as a string, in the form 'filename:lineno'.

  Args:
    default: Any = None, will be returned if the function's calling place could not be determiend
    backtrack_frames: int = 1, how many frames to go back in the stack, if set to 1, will return the calling frame's line number, if set to 0 will return this module's internal line number (so dont) but may be useful when building a function that itself makes use of get_lineno() in which case you'd need to make it 2 so your function doesn't just return internal line number
    additional_offset: int = 0, since you cannot easily add to the stringified int, you may set this parameter to offset the returned line number by some amount.
  """

  frame = inspect.currentframe()
  for _ in range(backtrack_frames):
    if frame is None:
      raise RuntimeError('Invalid frame.')
    frame = frame.f_back

  return f'{frame.f_code.co_filename}:{get_lineno(default=default, backtrack_frames=backtrack_frames+1)+additional_offset}'


def get_pyversion_str(precision=2) -> str:
  return '.'.join([str(x) for x in sys.version_info[:precision]])



from typing import Any

from .tcr_iterable import getattr_queue


def get_classname(obj: Any, *, default: str = 'UnknownClass') -> str:
  return getattr_queue(obj, '__class__.__name__', default=default)


def get_name_classname(obj: Any, *, default: str = 'UnknownClass') -> str:
  return getattr_queue(obj, '__name__', '__class__.__name__', default=default)


def get_qualname_classname(obj: Any, *, default: str = 'UnknownClass') -> str:
  return getattr_queue(obj, '__qualname__', '__name__', '__class__.__name__', default=default)



class Every:
  """Return True every N calls of this instance. Additionally you can look up the total number of calls using `.count` attribute.

  Example:
  ```py
  every = Every(10)
  for i in range(100):
    if every():
      print(i)
  ```
  """

  count: int
  every: int

  def __init__(self, every: int, *, start: int = 0):
    if every == 0:
      raise ZeroDivisionError('every cannot be zero due to imminent division')
    self.every = every
    self.count = start

  def __call__(self, every: int | None = None) -> bool:
    try:
      return self.count % (every if every is not None else self.every) == 0
    finally:
      self.count += 1



import datetime as dt
from functools import partial

from . import tcr_print as _m_print
from .tcr_int import hex


class BrainfuckCode(str):
  """Used in tcr.fmt_iterable() in order to format it as Brainfuck code.

  This will filter out all non-brainfuck characters: `>.<[+-],`.

  This can be used outside of fmt_iterable() to filter characters if needed.
  """

  def __new__(cls, value):
    return super().__new__(cls, ''.join(filter(lambda x: x in '>.<[+-],', value)))


class QuotelessString(str):
  """Used in tcr.fmt_iterable() in order to format it as a plain text, rather than as `"normal str"`."""


class HexInt(int):
  leading_zeroes: int

  def __new__(cls, value, leading_zeroes=None, **kwargs):
    return super().__new__(cls, value, **kwargs)

  def __init__(self, value, *, leading_zeroes=6, **_):
    self.leading_zeroes = leading_zeroes
    super().__init__()

  def __tcr_fmt__(self, fmt_iterable, **kwargs):
    return fmt_iterable(int(self), **{**kwargs, 'int_formatter': partial(hex, leading_zeroes=self.leading_zeroes)})


class UnreprableString(QuotelessString):
  def __repr__(self) -> str:
    return self


class UnixTimestampInt(int):
  """Add timezone by assinging the tz attribute after instantiation."""

  def __init__(self, *_, **__) -> None:
    self.tz = None
    super().__init__()

  def _get_as_seconds_unix(self) -> int:
    """Convert this unix timestamp from any type (seconds, miliseconds, microseconds) to the seconds type by dividing by 1000 each time if it's too large to be the seconds variant."""
    i = int(self)

    while i > 99_999_999_999:  # Convert any milisecond or smaller values to seconds
      i //= 1000

    return i

  def __tcr_fmt__(self, fmt_iterable, **kwargs):
    fromtimestamp_value = self._get_as_seconds_unix()

    while fromtimestamp_value > 99_999_999_999:  # Convert any milisecond or smaller values to seconds
      fromtimestamp_value //= 1000

    lhs = fmt_iterable(dt.datetime.fromtimestamp(fromtimestamp_value, tz=self.tz), **kwargs)
    rhs = fmt_iterable([int(self)], **kwargs)

    return f'{lhs} {rhs}'

  def to_datetime(self, tz: dt.tzinfo | None = None) -> dt.datetime:
    """Evaluate this unix timestamp into a datetime.datetime object."""
    return dt.datetime.fromtimestamp(self._get_as_seconds_unix(), tz=tz)

  def to_int(self):
    """Forget that this int is a unix timestamp - You may still use this object in place of an int."""
    return int(self)


class GayString(str):
  """Used in tcr.fmt_iterable() in order to format it as gay str. Happy pride :3"""  # noqa: D400


class AlwaysTruthyStr(str):
  """A string that always passes an if check no matter the contents."""

  def __bool__(self):
    return True



try:
  from .tcri_handler import ImGuiHandler
except ImportError:

  def ImGuiHandler(*_, **__):
    from ..src.tcr_console import console as __c

    msg1 = 'You need to install tcrutils with the [full] extra or [imgui] extra to use this module'
    msg2 = 'pip install tcrutils[imgui]'
    __c.error(msg1)
    __c.error(msg2)
    raise RuntimeError(f'{msg1}\n{msg2}')


from .tcri_utils import Every
from .types import tcri_types_imgui as imtypes
